/*
 * File: htext_api.c
 * Exposed functions
 */
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "htext.h"
#include "htext_private.h"

/**
 * Dummy callback
 */
void dummy_callback()
{
  return;
}

/* Types associated with the options */
static option_entry option_definitions[] = {
  {OT_STRING, (option_value)NULL}, /* HTEXTOP_SOURCEURL       */
  {OT_STRING, (option_value)NULL}, /* HTEXTOP_DESTINATIONURL  */
  {OT_INT,    (option_value)1   }, /* HTEXTOP_NUMBEROFSTREAMS */

  {OT_STRING, (option_value)NULL}, /* HTEXTOP_USERCERTIFICATE */
  {OT_STRING, (option_value)NULL}, /* HTEXTOP_USERPRIVKEY     */
  {OT_STRING, (option_value)NULL}, /* HTEXTOP_USERPRIVKEYPASS */
  {OT_INT,    (option_value)720},  /* HTEXTOP_PROXYLIFE       */

  {OT_STRING, (option_value)NULL}, /* HTEXTOP_CAPATH     */
  {OT_STRING, (option_value)NULL}, /* HTEXTOP_CAFILE     */
  {OT_INT,    (option_value)1},    /* HTEXTOP_VERIFYPEER */

  {OT_STRING, (option_value)"htext"}, /* HTEXTOP_CLIENTID */

  {OT_POINTER, (option_value)(void*)dummy_callback}, /* HTEXTOP_LOGCALLBACK */
  {OT_POINTER, (option_value)NULL}, /* HTEXTOP_LOGCALLBACK_DATA       */
  {OT_INT,     (option_value)0},    /* HTEXT_VERBOSITY         */

  {OT_INT,     (option_value)0},    /* HTEXT_BUFFERSIZE        */

  {OT_STRING,  (option_value)NULL}, /* HTEXT_DELEGATION_URL */
};

/**
 * Checks if url is a remote path
 * @return 1 if remote. 0 if local.
 */
static int is_remote(const char *url)
{
  return strncmp(url, "https:", 6) == 0 ||
         strncmp(url, "http:", 5) == 0;
}

/**
 * Initializes internals
 * @return 0 on success.
 */
int htext_global_init()
{
  if (htext_locking_setup() != 0)
    return 1;

  return curl_global_init(CURL_GLOBAL_ALL);
}

/**
 * Create a new handle
 * @return A new handle, or NULL if there is not enough memory
 */
htext_handle* htext_init()
{
  htext_handle *handle;
  int           i;

  /* Alloc */
  handle = calloc(1, sizeof(htext_handle));

  /* Copy defaults */
  handle->options = calloc(HTEXTOP_SENTINEL, sizeof(option_entry));
  for(i = 0; i < HTEXTOP_SENTINEL; ++i) {
    handle->options[i] = option_definitions[i];
    if (handle->options[i].type == OT_STRING && option_definitions[i].value.s)
      handle->options[i].value.s = strdup(option_definitions[i].value.s);
  }

  /* Initialize fields */
  handle->curl_share = curl_share_init();
  curl_share_setopt(handle->curl_share,
                    CURLSHOPT_SHARE,
                    CURL_LOCK_DATA_SSL_SESSION);

  /* Done */
  handle->status = HTEXTS_CREATED;
  return handle;
}


/**
 * Destroy the handle and free the associated memory
 * @param handle The handle to free
 */
void htext_destroy(htext_handle *handle)
{
  int i;

  if (handle->thread) pthread_join(handle->thread, NULL);

  switch (handle->status) {
    case HTEXTS_WAITING: case HTEXTS_STARTING:
    case HTEXTS_RUNNING:
      htext_abort(handle);
  }

  curl_share_cleanup(handle->curl_share);
  free(handle->error_string);
  free(handle->partial_total);
  free(handle->partial_done);

  for (i = 0; i < HTEXTOP_SENTINEL; ++i) {
    if (handle->options[i].type == OT_STRING)
      free(handle->options[i].value.s);
  }

  free(handle->options);
  free(handle);
}

/**
 * Set an option
 * @param handle The handle where to set the option
 * @param option The option (see HTEXT_OPTION)
 * @param ...    The value. Strings are duplicated, but pointers aren't
 * @return 0 on success. 1 on failure, and errno will be set.
 */
int htext_setopt(htext_handle *handle, unsigned option, ...)
{
  va_list       list;
  option_entry *op;
  const char   *c;

  /* Valid option? */
  if (option >= HTEXTOP_SENTINEL) {
    errno = EINVAL;
    return 1;
  }

  /* Handle in valid status? */
  switch (handle->status) {
    case HTEXTS_WAITING: case HTEXTS_STARTING:
    case HTEXTS_RUNNING:
      errno = EBADFD;
      return 1;
  }

  /* Set */
  op = &(handle->options[option]);

  va_start(list, option);
  switch (op->type) {
    case OT_STRING:
      c = va_arg(list, const char*);
      free(op->value.s);
      op->value.s = c?strdup(c):NULL;
      break;
    case OT_POINTER:
      op->value.p = va_arg(list, void*);
      break;
    case OT_INT:
      op->value.i = va_arg(list, int);
      break;
    default:
      errno = ENOSYS;
      return 1;
  }
  va_end(list);

  return 0;
}

/**
 * Perform the appropiate action. This means, put, get or copy
 * depending on the source and URL
 * @param handle
 * @return 0 on success. Otherwise, errno is set.
 */
int htext_perform(htext_handle *handle)
{
  void* (*performer)(void *);
  const char *performer_str;
  
  /* Check status */
  switch (handle->status) {
    case HTEXTS_WAITING: case HTEXTS_STARTING:
    case HTEXTS_RUNNING:
      errno = EBADFD;
      return 1;
  }

  /* Check we have source and destination */
  if (!GETSTR(handle, HTEXTOP_SOURCEURL)) {
    htext_error(handle, "Source not specified");
    errno = EINVAL;
    return 1;
  }
  if (!GETSTR(handle, HTEXTOP_DESTINATIONURL)) {
    htext_error(handle, "Destination not specified");
    errno = EINVAL;
    return 1;
  }

  /* Guess the method */
  if (is_remote(GETSTR(handle, HTEXTOP_SOURCEURL))) {
    if(is_remote(GETSTR(handle, HTEXTOP_DESTINATIONURL)))
      handle->http_method = HTTP_COPY;
    else
      handle->http_method = HTTP_GET;
  }
  else {
    if(is_remote(GETSTR(handle, HTEXTOP_DESTINATIONURL)))
      handle->http_method = HTTP_PUT;
    else
      handle->http_method = HTTP_UNSUPPORTED;
  }

  /* Delegate */
  switch (handle->http_method) {
    case HTTP_GET:
      performer = htext_get_method;
      performer_str = "GET";
      break;
    case HTTP_PUT:
      performer = htext_put_method;
      performer_str = "PUT";
      break;
    case HTTP_COPY:
      performer = htext_copy_method;
      performer_str = "COPY";
      break;
    default:
      errno = ENOSYS;
      return 1;
  }

  /* Launch thread and return */
  handle->status = HTEXTS_STARTING;
  htext_log(handle, "Delegating to the proper handle (%s)", performer_str);
  return pthread_create(&(handle->thread), NULL, performer, handle);
}

/**
 * Abort the transfer
 * @param handle The handle to abort
 */
void htext_abort(htext_handle *handle)
{
  /* TODO: Really abort */
  handle->status = HTEXTS_ABORTED;
}

/**
 * Return the status of the handle
 * @param handle
 * @return See HTEXT_STATUS
 */
int htext_status(htext_handle *handle)
{
  return handle->status;
}

/**
 * Put into total and done the current progress
 * @param handle The handle
 * @param n      Number of streams being used
 * @param total  A pointer to n partial totals will be set
 * @param done   A pointer to n partial dones will be set
 * @return 0 on success. Otherwise, errno is set.
 */
int htext_progress(htext_handle *handle, size_t *n, size_t **total, size_t **done)
{
  if (handle->status == HTEXTS_CREATED) {
    errno = EBADFD;
    return 1;
  }
  if (n)     *n     = handle->partials;
  if (done)  *done  = handle->partial_done;
  if (total) *total = handle->partial_total;
  return 0;
}

/**
 * Return the error string, if status is failed or aborted
 * @param handle
 * @return The error string, or NULL if not available
 */
const char *htext_error_string(htext_handle *handle)
{
  return handle->error_string;
}

/**
 * Return the HTTP return code. Useful in case an error happens.
 * @param handle
 * @return 0 if not set or the action failed before the server could be contacted.
 */
int htext_http_code (htext_handle *handle)
{
  return handle->http_status;
}
