/*
 * File: htext_copy.c
 * Implementation of the COPY method
 */
#include <gridsite.h>
#include <openssl/err.h>
#define WITH_OPENSSL /* Or we will get nice stack smashing */
#include <stdsoap2.h>
#include "soapH.h"
#include "htext.h"
#include "htext_private.h"


/**
 * Do the delegation
 * @param handle     The handle
 * @param url        The delegation endpoint
 * @param err_buffer Used to build the error string
 * @return NULL on failure, or an allocated string with the delegation ID
 */
static char *htext_delegate(htext_handle *handle, char *url, char *err_buffer)
{
  char                               *delegation_id = NULL;
  char                               *reqtxt  = NULL, *certtxt = NULL;
  char                               *keycert = NULL;
  struct soap                         soap_get, soap_put;
  struct tns__getNewProxyReqResponse  getNewProxyReqResponse;
  char                               *ucert, *ukey, *capath;
  int                                 lifetime;

  /* Get from the handle */
  ucert    = GETSTR(handle, HTEXTOP_USERCERTIFICATE);
  ukey     = GETSTR(handle, HTEXTOP_USERPRIVKEY);
  capath   = GETSTR(handle, HTEXTOP_CAPATH);
  lifetime = GETINT(handle, HTEXTOP_PROXYLIFE);

  /* Only one is needed if they are the same */
  if (ucert && !ukey) ukey  = ucert;
  if (!ucert && ukey) ucert = ukey;

  /* Cert and key need to be in the same file */
  if (strcmp(ucert, ukey) == 0) {
    keycert = strdup(ucert);
  }
  else {
    FILE *ifp, *ofp;
    int   fd;
    char  c;

    keycert = strdup("/tmp/.XXXXXX");

    fd = mkstemp(keycert);
    ofp = fdopen(fd, "w");

    ifp = fopen(ukey, "r");
    while ((c = fgetc(ifp)) != EOF) fputc(c, ofp);
    fclose(ifp);

    ifp = fopen(ukey, "r");
    while ((c = fgetc(ifp)) != EOF) fputc(c, ofp);
    fclose(ifp);

    fclose(ofp);
  }

  /* Initialize SSL */
  ERR_load_crypto_strings ();
  OpenSSL_add_all_algorithms();

  /* Request a new delegation ID */
  soap_init(&soap_get);

  if (soap_ssl_client_context(&soap_get, SOAP_SSL_DEFAULT, keycert, "",
                              NULL, capath, NULL) == 0) {
    soap_call_tns__getNewProxyReq(&soap_get,
                                  url,
                                  "http://www.gridsite.org/namespaces/delegation-1",
                                  &getNewProxyReqResponse);

    if(soap_get.error == 0) {
      reqtxt        = getNewProxyReqResponse.getNewProxyReqReturn->proxyRequest;
      delegation_id = strdup(getNewProxyReqResponse.getNewProxyReqReturn->delegationID);

      /* Generate proxy */
      if (GRSTx509MakeProxyCert(&certtxt, stderr, reqtxt,
                                ucert, ukey, lifetime) == GRST_RET_OK) {
        /* Submit the proxy */
        soap_init(&soap_put);

        if (soap_ssl_client_context(&soap_put, SOAP_SSL_DEFAULT, keycert, "",
                              NULL, capath, NULL) == 0) {
            soap_call_tns__putProxy(&soap_put,
                          url,
                          "http://www.gridsite.org/namespaces/delegation-1",
                          delegation_id, certtxt, NULL);
            if (soap_put.error) {
              /* Could not PUT */
#ifndef NO_SOAP_SPRINT
              soap_sprint_fault(&soap_put, err_buffer, sizeof(err_buffer));
              htext_error(handle, err_buffer);
#else
              soap_print_fault(&soap_put, stderr);
              handle->status = HTEXTS_FAILED;
#endif
            }
        }
        else { /* soap_put ssl error */
#ifndef NO_SOAP_SPRINT
          soap_sprint_fault(&soap_put, err_buffer, sizeof(err_buffer));
          htext_error(handle, err_buffer);
#else
          soap_print_fault(&soap_put, stderr);
          handle->status = HTEXTS_FAILED;
#endif
        }

        soap_end(&soap_put);
        soap_done(&soap_put);
      }
      else {
        htext_error(handle, "Could not generate the proxy");
      }
    }
    else { /* Could not get ID */
#ifndef NO_SOAP_SPRINT
      soap_sprint_fault(&soap_get, err_buffer, sizeof(err_buffer));
      htext_error(handle, err_buffer);
#else
      soap_print_fault(&soap_put, stderr);
      handle->status = HTEXTS_FAILED;
#endif
    }
  }
  else { /* soap_get ssl error */
#ifndef NO_SOAP_SPRINT
    soap_sprint_fault(&soap_get, err_buffer, sizeof(err_buffer));
    htext_error(handle, err_buffer);
#else
    soap_print_fault(&soap_put, stderr);
    handle->status = HTEXTS_FAILED;
#endif
  }

  /* Clean soap_get */
  soap_end(&soap_get);
  soap_done(&soap_get);
  free(keycert);
  free(certtxt);
  
  /* Return delegation ID */
  return delegation_id;
}

/**
 * Used to parse the output from the server to know the progress
 */
size_t htext_copy_write_callback(char *buffer, size_t size, size_t nmemb, void *pp)
{
  htext_partial *partial = (htext_partial*)pp;
  size_t         received = size * nmemb;
  size_t         bsize, parsed;
  size_t         aux_done, aux_total;
  char          *next, *prev;
  char          *remaining = (char*)partial->extra;

  htext_write_callback(buffer, size, nmemb, pp);


  if (remaining[0] != '\0') {
    size_t r_size = strlen(remaining);
    memcpy(remaining + r_size, buffer, received);
    bsize = received + r_size;
    memcpy(buffer, remaining, bsize);
    remaining[0] = '\0';
  }
  else
    bsize = received;

  buffer[bsize] = '\0'; /* To prevent strchr from going beyond */

  next   = buffer;
  parsed = 0;
  while (parsed < bsize) {
    if (sscanf(next, "%zu/%zu\n", &aux_done, &aux_total) == 2) {
      prev = next;
      next = strchr(next, '\n');
      /* Maybe sscanf got both numbers, but the line was incomplete!*/
      if (next == NULL) {
        memcpy(remaining, prev, bsize - parsed);
        remaining[bsize - parsed] = '\0';
        parsed = bsize;
      }
      else {
        *partial->partial_done  = aux_done;
        *partial->partial_total = aux_total;
        ++next;
        parsed = next - buffer;
      }
    }
    else {
      /* If it doesn't follow, it may be a message */
      if (strncasecmp("success", next, 7) == 0) {
        // Leave things going
      }
      else if (strncasecmp("aborted", next, 7) == 0) {
        partial->handle->status = HTEXTS_ABORTED;
      }
      else if (strncasecmp("failed", next, 6) == 0) {
        htext_error(partial->handle, next);
      }
      else {
        memcpy(remaining, next, bsize - parsed);
        remaining[bsize - parsed] = '\0';
      }
      parsed = bsize;
    }
  }

  return received;
}

/**
 * Performs the COPY
 * @param h The handle
 * @return NULL when done
 */
void *htext_copy_method(void *h)
{
  htext_handle    *handle = (htext_handle*)h;
  CURL            *curl;
  char            *delegation_id = NULL, *location;
  htext_partial    control;
  char             buffer[CURL_MAX_WRITE_SIZE], err_buffer[CURL_ERROR_SIZE];
  char             host[64];

  /* Create and initialize CURL handle */
  curl = curl_easy_init();

  curl_easy_setopt(curl, CURLOPT_ERRORBUFFER,      err_buffer);
  curl_easy_setopt(curl, CURLOPT_HEADERFUNCTION,   htext_header_callback);
  curl_easy_setopt(curl, CURLOPT_NOPROGRESS,       1);
  curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION,    htext_write_callback);
  curl_easy_setopt(curl, CURLOPT_WRITEDATA,        &control);
  curl_easy_setopt(curl, CURLOPT_URL,              GETSTR(handle, HTEXTOP_SOURCEURL));
  curl_easy_setopt(curl, CURLOPT_USERAGENT,        GETSTR(handle, HTEXTOP_CLIENTID));
  curl_easy_setopt(curl, CURLOPT_CAPATH,           GETSTR(handle, HTEXTOP_CAPATH));
  curl_easy_setopt(curl, CURLOPT_SSLCERT,          GETSTR(handle, HTEXTOP_USERCERTIFICATE));
  curl_easy_setopt(curl, CURLOPT_SSLKEY,           GETSTR(handle, HTEXTOP_USERPRIVKEY));
  curl_easy_setopt(curl, CURLOPT_SSLKEYPASSWD,     GETSTR(handle, HTEXTOP_USERPRIVKEYPASS));

  if (GETINT(handle, HTEXTOP_VERBOSITY) > 1) {
    curl_easy_setopt(curl, CURLOPT_DEBUGFUNCTION, htext_debug_callback);
    curl_easy_setopt(curl, CURLOPT_VERBOSE,       1);
  }

  curl_easy_setopt(curl, CURLOPT_SHARE, handle->curl_share);

  /* Reserve space for the partials */
  handle->partial_done  = calloc(sizeof(size_t), 1);
  handle->partial_total = calloc(sizeof(size_t), 1);
  handle->partials      = 1;

  /* Do a HEAD to know where do we have to go */
  htext_partial_init(&control);
  control.partial_total = handle->partial_total;
  control.partial_done  = handle->partial_done;
  control.handle        = handle;

  curl_easy_setopt(curl, CURLOPT_NOBODY,         1);
  curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1);
  curl_easy_setopt(curl, CURLOPT_HEADERDATA,     &control);
  curl_easy_setopt(curl, CURLOPT_DEBUGDATA,      &control);

  handle->status = HTEXTS_WAITING;
  htext_log(handle, "Retrieving the actual location");
  if (curl_easy_perform(curl) != CURLE_OK)
    htext_error(handle, buffer);

  /* If we got it, delegate */
  if (handle->status != HTEXTS_FAILED && control.location)
    location = control.location;
  else
    htext_error(handle, "Could not get the location");

  /* Delegate */
  if (handle->status != HTEXTS_FAILED) {
    /* Pick from configuration is not available */
    if (control.delegation_service == NULL) {
      control.delegation_service = GETSTR(handle, HTEXTOP_DELEGATION_URL);
    }
    /* Not yet? */
    if (control.delegation_service == NULL) {
      htext_error(handle, "Could not get the delegation endpoint");
    }
    /* Build the full URL */
    else {
      snprintf(buffer, sizeof(buffer), "https://%s%s",
                                       get_host(location, host),
                                       control.delegation_service);
      htext_log(handle, "Delegation endpoint: %s", buffer);
      delegation_id = htext_delegate(handle, buffer, err_buffer);
      htext_log(handle, "Got delegation id %s", delegation_id);
    }
  }

  /* Do the COPY */
  if (handle->status != HTEXTS_FAILED) {
    char *full_url;
    
    htext_log(handle, "Running the remote copy");
    handle->status = HTEXTS_RUNNING;

    full_url = malloc(strlen(location) + strlen(delegation_id) + 13);
    strcpy(full_url, location);
    strcat(full_url, "&delegation=");
    strcat(full_url, delegation_id);

    curl_easy_setopt(curl, CURLOPT_HTTPGET,       1);
    curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "COPY");
    curl_easy_setopt(curl, CURLOPT_URL,           full_url);
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, htext_copy_write_callback);
    curl_easy_setopt(curl, CURLOPT_WRITEDATA,     &control);
    /* Need to flush often to get progress */
    curl_easy_setopt(curl, CURLOPT_BUFFERSIZE,    64);

    snprintf(buffer, sizeof(buffer),
             "Destination: %s", GETSTR(handle, HTEXTOP_DESTINATIONURL));
    control.headers = curl_slist_append(control.headers, buffer);
    curl_easy_setopt(curl, CURLOPT_HTTPHEADER, control.headers);

    control.http_status = 0; /* Reset this */
    control.extra = buffer;
    buffer[0] = '\0';

    if (curl_easy_perform(curl) != CURLE_OK) {
      htext_error(handle, err_buffer);
    }
    else if (control.http_status >= 400 || handle->status == HTEXTS_FAILED) {
      htext_error(handle, NULL);
    }
    else {
      handle->status = HTEXTS_SUCCEEDED;
    }

    handle->http_status = control.http_status;

    free(full_url);
  }

  /* Clean up */
  free(delegation_id);
  htext_partial_clean(&control);
  curl_easy_cleanup(curl);
  return NULL;
}
