/*
 * File: locks.c
 * Locking implementation for OpenSSL and GnuTLS
 */
#ifdef WITH_GNUTLS
#include <gnutls/gnutls.h>
#endif

#ifndef WITHOUT_OPENSSL
#include <openssl/crypto.h>
#include <pthread.h>

static pthread_mutex_t *lock_array;
static int              n_locks;

void openSSL_lock(int mode, int n, const char *file, int line)
{
  if (mode & CRYPTO_LOCK) {
    pthread_mutex_lock(&(lock_array[n]));
  }
  else {
    pthread_mutex_unlock(&(lock_array[n]));
  }
}

#endif

unsigned long openSSL_thread_id(void)
{
  return (unsigned long)pthread_self();
}

/**
 * Initializes locking for OpenSSL/GnuTLS
 * @return 0 on success.
 */
int htext_locking_setup(void)
{
  static int mutexes_set = 0;
#ifdef WITH_GNUTLS
  if (gnutls_global_init() != 0)
    return 1;
#endif
#ifndef WITHOUT_OPENSSL
    /* Initialize mutexes */
  if (!mutexes_set) {
    int i;

    n_locks = CRYPTO_num_locks();

    /* Won't be freed, but that's fine, as it must survive as long as the process does */
    lock_array = (pthread_mutex_t*)OPENSSL_malloc(n_locks * sizeof(pthread_mutex_t));

    for (i = 0; i < n_locks; ++i) {
      pthread_mutex_init(&(lock_array[i]), NULL);
    }

    CRYPTO_set_id_callback(openSSL_thread_id);
    CRYPTO_set_locking_callback(openSSL_lock);
    
    mutexes_set = 1;
  }
#endif
  return 0;
}
