#include "../mod_dav/mod_dav.h"
#include "mod_dav_dpm.h"
#include "mod_ssl-private.h"
#include <apr_strings.h>
#include <apr_uri.h>
#include <ctype.h>
#include <curl/curl.h>
#include <http_log.h>
#include <http_protocol.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include "../client/htext.h"

typedef struct {
  apr_bucket_brigade *brigade;
  ap_filter_t        *output;
  request_rec        *request;
} dav_dpm_copy_data;

/** Return a pointer to a malloc'd string holding a URL-encoded (RFC 1738)
 * version of *in. Only A-Z a-z 0-9 . _ - are passed through unmodified.
 * (DN's processed by GRSThttpUrlEncode can be used as valid Unix filenames,
 * assuming they do not exceed restrictions on filename length.)
 * Adapted from GridSite
 */
static char *GRSThttpUrlEncode(apr_pool_t *pool, char *in)
{
  char *out, *p, *q;
  int   outlen;

  outlen = 3 * strlen(in) + 1;
  out = apr_pcalloc(pool, outlen);

  p = in;
  q = out;

  while (*p != '\0') {
   if (isalnum(*p) || (*p == '.') || (*p == '_') || (*p == '-')) {
    *q = *p;
    ++q;
   }
   else if (*p == ' ') {
     /* Spaces are replaced by + by GridSite */
     snprintf(q, outlen, "%%2B");
     q = &q[3];
   }
   else {
     snprintf(q, outlen, "%%%2X", *p);
     q = &q[3];
   }
   ++p;
  }

  *q = '\0';
  return out;
}

/**
 * Called by htext when there is logging output
 * @param handle The handle that triggers the call
 * @param type   Type of logging data
 * @param msg    The message itself
 * @param size   The size of the message (it might not be NULL terminated)
 * @param ud     User defined data
 */
static void dav_dpm_copy_log(htext_handle *handle, HTEXT_LOG_TYPE type, const char *msg, size_t size, void *ud)
{
  dav_dpm_copy_data *ddc = (dav_dpm_copy_data*)ud;

  switch (type) {
   case HTEXT_LOG:
     ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, ddc->request,
                   "%s", msg);
     break;
    default:
      break;
 }
}

/**
 * Performs a remote copy
 * @param src    The original resource
 * @param remote The remote URL
 * @param depth  The copy depth
 * @return       NULL on success
 */
dav_error *dav_dpm_remote_copy(const dav_resource *src, const char *remote,
                               int depth)
{
  apr_uri_t           remote_uri;
  dav_error          *error = NULL;
  apr_bucket         *bkt;
  apr_bucket_brigade *bb;
  const char         *delegation, *error_string;
  char               *uproxy;
  struct stat         fstat;
  SSLSrvConfigRec    *ssl_conf;
  htext_handle       *handle;
  int                 wait, status;
  dav_dpm_copy_data   ddc;
  struct timespec     sleep_time = {0, 5000};

  /* mod_ssl configuration */
  ssl_conf = (SSLSrvConfigRec*)ap_get_module_config(src->info->request->server->module_config,
                                                    &ssl_module);

  /* Parse URI */
  apr_uri_parse(src->pool, remote, &remote_uri);

  /* Local copies are not accepted */
  if (apr_strnatcasecmp(remote_uri.hostname,
                        src->info->request->server->server_hostname) == 0)
  {
    return dav_dpm_new_error(src->pool, EPERM, HTTP_METHOD_NOT_ALLOWED,
                             "Local copies not allowed");
  }

  /* Check permissions */
  if (!(src->info->d_conf->flags & DAV_DPM_REMOTE_COPY)) {
    return dav_dpm_new_error(src->pool, EPERM, HTTP_METHOD_NOT_ALLOWED,
                             "Remote copies not allowed");
  }

  /* Get delegation proxy */
  if (!(delegation = apr_table_get(src->info->query, "delegation"))) {
    return dav_dpm_new_error(src->pool, EINVAL, HTTP_BAD_REQUEST,
                             "A delegation id is needed for remote copies");
  }

  uproxy = apr_pstrcat(src->pool,
                       src->info->d_conf->proxy_cache,
                       "/%3A",
                       GRSThttpUrlEncode(src->pool, src->info->user->principal),
                       "/", delegation, "/userproxy.pem",
                       NULL
                      );
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, src->info->request,
                "Using delegation proxy '%s'", uproxy);

  if (stat(uproxy, &fstat) != 0) {
    return dav_dpm_new_error(src->pool, EACCES, 0,
                             "Could not stat %s", uproxy);
  }

  if (!S_ISREG(fstat.st_mode)) {
    return dav_dpm_new_error(src->pool, EACCES, 0,
                             "The delegated proxy does not exist: %s", uproxy);
  }


  /* Get the physical location */
  if (!src->info->location.token) {
    if ((error = dav_dpm_get_physical_location(src->info, src->pool)))
      return error;
  }

  /* Redirect? */
  if (src->info->redirect) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, src->info->request,
                  "Resource %s located at %s",
                   src->info->sfn, src->info->redirect);

    apr_table_set(src->info->request->headers_out, "Location", src->info->redirect);
    return dav_new_error(src->pool, HTTP_TEMPORARY_REDIRECT, 0, src->info->redirect);
  }

  /* Setup the handle */
  if (!(handle = htext_init())) {
    return dav_dpm_new_error(src->pool, ENOMEM, 0,
                             "Not enough memory");
  }

  htext_setopt(handle, HTEXTOP_SOURCEURL,       src->info->location.path);
  htext_setopt(handle, HTEXTOP_DESTINATIONURL,  remote);

  htext_setopt(handle, HTEXTOP_USERCERTIFICATE, uproxy);
  htext_setopt(handle, HTEXTOP_USERPRIVKEY,     uproxy);

  htext_setopt(handle, HTEXTOP_CAPATH,     ssl_conf->server->auth.ca_cert_path);
  htext_setopt(handle, HTEXTOP_CAFILE,     ssl_conf->server->auth.ca_cert_file);
  htext_setopt(handle, HTEXTOP_VERIFYPEER, 1);

  ddc.output  = src->info->request->output_filters;
  ddc.brigade = apr_brigade_create(src->pool, ddc.output->c->bucket_alloc);
  ddc.request = src->info->request;
  htext_setopt(handle, HTEXTOP_LOGCALLBACK,      dav_dpm_copy_log);
  htext_setopt(handle, HTEXTOP_LOGCALLBACK_DATA, &ddc);
  htext_setopt(handle, HTEXTOP_VERBOSITY,        1);

  /* Run */
  if (htext_perform(handle) != 0) {
    htext_destroy(handle);
    return dav_dpm_new_error(src->pool, errno, 0,
                             "Could not perform the action");
  }

  /* Wait */
  src->info->request->status = 0;
  bb  = apr_brigade_create(src->pool, ddc.output->c->bucket_alloc);
  bkt = apr_bucket_flush_create(ddc.output->c->bucket_alloc);
  do {
    size_t *total, *done, n, i;
    size_t  globalDone, globalTotal;

    status = htext_status(handle);
    switch (status) {
      case HTEXTS_SUCCEEDED: case HTEXTS_FAILED:
      case HTEXTS_ABORTED:
        wait = 0;
        break;
      default:
        /* In the first go we need to set the reply */
        if (src->info->request->status == 0) {
          src->info->request->status = HTTP_ACCEPTED;
          ap_set_content_type(src->info->request, "text/plain");
        }
        /* Print progress */
        htext_progress(handle, &n, &total, &done);
        globalDone = globalTotal = 0;
        for (i = 0; i < n; ++i) {
          globalDone  += done[i];
          globalTotal += total[i];
        }                           
        apr_brigade_printf(ddc.brigade, ap_filter_flush, ddc.output,
                           "%lu/%lu\n",
                           (unsigned long)globalDone, (unsigned long)globalTotal);
        ap_fflush(ddc.output, ddc.brigade);
        /* Sleep a while */
        wait = 1;
        nanosleep(&sleep_time, NULL);
    }
  } while (wait);

  /* Feedback */
  error_string = htext_error_string(handle);
  switch (status) {
    case HTEXTS_SUCCEEDED:
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, src->info->request,
                    "Remote copy finished successfully (%d): %s => %s",
                    htext_http_code(handle), src->uri, remote);
      /* 200 is OK, but 201 fits better */
      if (src->info->request->status == 0)
        src->info->request->status = HTTP_CREATED;
      apr_brigade_printf(ddc.brigade, ap_filter_flush, ddc.output, "Success\n");
      break;
    case HTEXTS_FAILED:
      if (src->info->request->status == 0)
        error = dav_dpm_new_error(src->pool, EIO, htext_http_code(handle),
                                  "Failed: Remote copy failed with status code %d: %s",
                                  htext_http_code(handle),
                                  error_string?error_string:"");
      else
        apr_brigade_printf(ddc.brigade, ap_filter_flush, ddc.output,
                           "Failed: Remote copy failed with status code %d: %s\n",
                           htext_http_code(handle),
                           error_string?error_string:"");
      break;
    case HTEXTS_ABORTED:
      if (src->info->request->status == 0)
        error = dav_dpm_new_error(src->pool, EIO, HTTP_INTERNAL_SERVER_ERROR,
                                  "Aborted");
      else
        apr_brigade_printf(ddc.brigade, ap_filter_flush, ddc.output, "Aborted\n");
      break;
  }

  /* Finish */
  htext_destroy(handle);

  if (!error) {
    bkt = apr_bucket_eos_create(ddc.output->c->bucket_alloc);
    APR_BRIGADE_INSERT_TAIL(ddc.brigade, bkt);
    if (ap_pass_brigade(ddc.output, ddc.brigade) != APR_SUCCESS)
      error = dav_dpm_new_error(src->pool, EIO, 0, "Could not write EOS to filter.");
  }

  return error;
}

