#include "mod_dav_dpm.h"
#include <apr_strings.h>
#include <pthread.h>
#include "../client/htext.h"

/* Provider definition */
static const dav_provider dav_dpm_provider =
{
    &dav_hooks_repository_dpm, /* Repository handlers  */
    &dav_hooks_db_dbm,         /* DB Handlers          */
    NULL,                      /* Lock                 */
    NULL,                      /* Versioning           */
    NULL,                      /* Binding              */
    NULL,                      /* Search               */
    NULL                       /* Context              */
};

/* Correspondence string->flag */
typedef struct {
  const char *str;
  dir_flags   flag;
} dav_dpm_dir_flags;

static dav_dpm_dir_flags dav_dpm_dir_flags_known[] = {
  {"write"  ,         DAV_DPM_WRITE},
  {"diskdirallow",    DAV_DPM_DISK_DIR_ALLOW},
  {"diskdirredirect", DAV_DPM_DISK_DIR_REDIRECT},
  {"remotecopy",      DAV_DPM_REMOTE_COPY},
  {NULL}
};

/* Prototypes */
int Cthread_init(void);
int _Cthread_addcid(char *, int, char *, int, pthread_t, unsigned, void *(*)(void *), int);

/**
 * Creates a structure for the server configuration
 * @param p The pool used to allocate memory
 * @param s Information about the virtual server being configured
 * @return A clean dav_dpm_server_conf instance
 */
static void *dav_dpm_create_server_config(apr_pool_t *p, server_rec *s)
{
  dav_dpm_server_conf *conf = apr_pcalloc(p, sizeof(dav_dpm_server_conf));

  /* Set defaults */
  conf->csec_mech = "ID";
  setenv("CSEC_MECH", conf->csec_mech, 1);
  conf->dav_host = conf->dpm_host = conf->ns_host = NULL;
  conf->type     = DAV_DPM_NODE_HEAD;

  return conf;
}

/**
 * Creates a structure for the directory configuration
 * @param p   The pool used to allocate memoryu
 * @param dir The directory being cofigured
 * @return A clean dav_dpm_dir_conf instance
 */
static void *dav_dpm_create_dir_config(apr_pool_t *p, char *dir)
{
  dav_dpm_dir_conf *conf = apr_pcalloc(p, sizeof(dav_dpm_dir_conf));

  /* Set defaults */
  conf->base         = "";
  conf->flags        = 0x00;
  conf->proxy_cache  = "/var/proxycache";
  conf->filesystems  = apr_hash_make(p);

  conf->generate_token = dpm_generate_token;
  conf->validate_token = dpm_validate_token;
  conf->token_extra    = NULL;

  return conf;
}

/**
 * Function called for the configuration parameter NSHost
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config Unused pointer
 * @param arg    The value of the parameter
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_ns_host(cmd_parms *cmd, void *config, const char *arg)
{
  dav_dpm_server_conf *conf;
  conf = ap_get_module_config(cmd->server->module_config, &dav_dpm_module);
  conf->ns_host = arg;
  if (!conf->ns_host)
    return apr_pstrcat(cmd->pool, "Invalid NSHost ", arg, NULL);
  if (setenv("DPNS_HOST", conf->ns_host, 1))
    return apr_pstrcat(cmd->pool, "Could not set DPNS_HOST (", arg, ")", NULL);
  if (setenv("LFC_HOST", conf->ns_host, 1))
    return apr_pstrcat(cmd->pool, "Could not set LFC_HOST (", arg, ")", NULL);

  return NULL;
}

/**
 * Function called for the configuration parameter DPMHost
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config Unused pointer
 * @param arg    The value of the parameter
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_dpm_host(cmd_parms *cmd, void *config, const char *arg)
{
  dav_dpm_server_conf *conf;
  conf = ap_get_module_config(cmd->server->module_config, &dav_dpm_module);
  conf->dpm_host = arg;
  if (!conf->dpm_host)
    return apr_pstrcat(cmd->pool, "Invalid DPMHost ", arg, NULL);
  if (setenv("DPM_HOST", conf->dpm_host, 1))
    return apr_pstrcat(cmd->pool, "Could not set DPM_HOST (", arg, ")", NULL);

  if (conf->dav_host == NULL)
    conf->dav_host = conf->dpm_host;
  if (conf->ns_host == NULL)
    return dav_dpm_cmd_ns_host(cmd, config, arg);

  return NULL;
}

/**
 * Function called for the configuration parameter DAVHost
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config Unused pointer
 * @param arg    The value of the parameter
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_dav_host(cmd_parms *cmd, void *config, const char *arg)
{
  dav_dpm_server_conf *conf;
  conf = ap_get_module_config(cmd->server->module_config, &dav_dpm_module);
  conf->dav_host = arg;
  if (!conf->dav_host)
    return apr_pstrcat(cmd->pool, "Invalid DAVHost ", arg, NULL);
  return NULL;
}

/**
 * Function called for the configuration parameter NodeType
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config Unused pointer
 * @param arg    The value of the parameter
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_node_type(cmd_parms *cmd, void *config, const char *arg)
{
  dav_dpm_server_conf *conf;
  conf = ap_get_module_config(cmd->server->module_config, &dav_dpm_module);
  if (strcasecmp(arg, "head") == 0)
    conf->type = DAV_DPM_NODE_HEAD;
  else if (strcasecmp(arg, "disk") == 0)
    conf->type = DAV_DPM_NODE_DISK;
  else if (strcasecmp(arg, "lfc") == 0)
    conf->type = DAV_DPM_NODE_LFC;
  else
    return "Unknown value for NodeType";
  return NULL;
}

/**
 * Function called for the configuration parameter CSEC_MECH
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config Unused pointer
 * @param arg    The value of the parameter
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_csec_mech(cmd_parms *cmd, void *config, const char *arg)
{
  dav_dpm_server_conf *conf;
  conf = ap_get_module_config(cmd->server->module_config, &dav_dpm_module);
  conf->csec_mech = arg;
  if (!conf->csec_mech)
    return apr_pstrcat(cmd->pool, "Invalid CSEC_MECH ", arg, NULL);
  if (setenv("CSEC_MECH", conf->csec_mech, 1))
    return apr_pstrcat(cmd->pool, "Could not set CSEC_MECH (", arg, ")", NULL);
  return NULL;
}

/**
 * Function called for the configuration parameter UserMappingCache
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config Unused pointer
 * @param arg    The value of the parameter
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_umap_cache(cmd_parms *cmd, void *config, const char *file, const char *size)
{
  dav_dpm_server_conf *conf;
  apr_status_t         stat;
  int                  nentries;

  conf = ap_get_module_config(cmd->server->module_config, &dav_dpm_module);

  /* Remove if exists */
  stat = apr_file_remove(file, cmd->pool);
  if (stat != APR_SUCCESS && stat != ENOENT)
    return apr_psprintf(cmd->pool, "Could not remove the shared memory file (errno %d)", stat);

  /* Create */
  nentries = atoi(size);
  if (nentries < 1)
    return "The size of the user map cache has to be at least 1";

  stat = apr_shm_create(&(conf->session_cache_shm), sizeof(dav_dpm_user_map_cache_entry) * nentries,
                        file, cmd->pool);

  if (stat != APR_SUCCESS)
    return apr_psprintf(cmd->pool, "Could not set the shared memory file (errno %d)", stat);

  conf->session_cache_size = nentries;

  return NULL;
}

/**
 * Function called for the configuration parameter DMBase (per directory)
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config A pointer to the directory configuration
 * @param arg    The value of the parameter
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_base(cmd_parms *cmd, void *config, const char *arg)
{
  dav_dpm_dir_conf *conf = (dav_dpm_dir_conf*)config;
  char             *aux;
  int len;
  /* Remove end "/", if any */
  len = strlen(arg);
  if (len > 0 && arg[len - 1] == '/') {
    aux          = apr_pstrdup(cmd->pool, arg);
    aux[len - 1] = '\0';
    conf->base   = aux;
  }
  else {
    conf->base = arg;
  }
  return NULL;
}

/**
 * Set the flags for a directory/location. Called once per flag.
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config A pointer to the directory configuration
 * @param w      The flag
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_flags(cmd_parms *cmd, void *config, const char *w)
{
  dav_dpm_dir_conf  *conf = (dav_dpm_dir_conf*)config;
  dav_dpm_dir_flags *iter;

  for(iter = dav_dpm_dir_flags_known; iter->str; ++iter) {
    if (strcasecmp(w, iter->str) == 0) {
      conf->flags |= iter->flag;
      return NULL;
    }
  }

  return apr_psprintf(cmd->pool, "%s is not a recognised flag", w);
}

/**
 * Set the "anonymous" user
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config A pointer to the directory configuration
 * @param arg    The user
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_anon(cmd_parms *cmd, void *config, const char *arg)
{
  dav_dpm_dir_conf *conf = (dav_dpm_dir_conf*)config;
  conf->anon_user = arg;
  return NULL;
}

/**
 * Set the proxy cache location
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config A pointer to the directory configuration
 * @param arg    The user
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_proxy_cache(cmd_parms *cmd, void *config, const char *arg)
{
  dav_dpm_dir_conf *conf = (dav_dpm_dir_conf*)config;
  conf->proxy_cache = arg;
  return NULL;
}

/**
 * Set the type of token used
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config A pointer to the directory configuration
 * @param method The method used
 * @param extra  A parameter for the method
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_token(cmd_parms *cmd, void *config, const char *method, const char *extra)
{
  dav_dpm_dir_conf *conf = (dav_dpm_dir_conf*)config;

  /* DPM token implementation */
  if (strcasecmp(method, "dpm") == 0) {
    conf->generate_token = dpm_generate_token;
    conf->validate_token = dpm_validate_token;
  }
  else if (strcasecmp(method, "password") == 0) {
    conf->generate_token = passwd_generate_token;
    conf->validate_token = passwd_validate_token;
    conf->token_extra    = extra;
    if (!extra)
      return "A password must be specified for DMToken password!";
  }
  else if (strcasecmp(method, "passfile") == 0) {
    conf->generate_token = file_generate_token;
    conf->validate_token = file_validate_token;
    conf->token_extra    = extra;
    if (!extra)
      return "A file must be specified for DMToken passfile!";
  }
  else {
    return "Invalid method!";
  }

  return NULL;
}

/**
 * Sets which subset of FileSystems (Paths) are accessible. This way, if someone
 * manages to break a token, still will be "trapped" here
 * @param cmd    Lots of information about the configuration contexts (as pool)
 * @param config A pointer to the directory configuration
 * @param w      The allowed path
 * @return       NULL on success. An error string otherwise
 */
static const char *dav_dpm_cmd_filesystems(cmd_parms *cmd, void *config, const char *w)
{
  dav_dpm_dir_conf *conf = (dav_dpm_dir_conf*)config;
  apr_hash_set(conf->filesystems, w, strlen(w), (void*)1);
  return NULL;
}

/**
 * Called to initialize each children
 * @param pool
 * @param server
 */
static void dav_dpm_child_init(apr_pool_t *pool, server_rec *server)
{
  Cthread_init();
  _Cthread_addcid(NULL, 0, NULL, 0, pthread_self(), 0, NULL, 0);
}

/* Command list (configuration parameters) */
static const command_rec dav_dpm_cmds[] =
{
  AP_INIT_TAKE1  ("DPMHost",       dav_dpm_cmd_dpm_host,    NULL, RSRC_CONF,   "DPM host"),
  AP_INIT_TAKE1  ("NSHost",        dav_dpm_cmd_ns_host,     NULL, RSRC_CONF,   "DPNS/LFC host"),
  AP_INIT_TAKE1  ("DAVHost",       dav_dpm_cmd_dav_host,    NULL, RSRC_CONF,   "DAV host"),
  AP_INIT_TAKE1  ("NodeType",      dav_dpm_cmd_node_type,   NULL, RSRC_CONF,   "Node type"),
  AP_INIT_TAKE1  ("CSEC_MECH",     dav_dpm_cmd_csec_mech,   NULL, RSRC_CONF,   "CSEC authentication method"),
  AP_INIT_TAKE2  ("UserMapCache",  dav_dpm_cmd_umap_cache,  NULL, RSRC_CONF,   "File used for shared memory"),
  AP_INIT_TAKE1  ("DMBase",        dav_dpm_cmd_base,        NULL, ACCESS_CONF, "Base directory"),
  AP_INIT_ITERATE("DMFlags",       dav_dpm_cmd_flags,       NULL, ACCESS_CONF, "Directory flags"),
  AP_INIT_TAKE1  ("DMAnon",        dav_dpm_cmd_anon,        NULL, ACCESS_CONF, "Anonymous user for fallback"),
  AP_INIT_TAKE1  ("ProxyCache",    dav_dpm_cmd_proxy_cache, NULL, ACCESS_CONF, "Proxy cache location"),
  AP_INIT_TAKE12 ("DMToken",       dav_dpm_cmd_token,       NULL, ACCESS_CONF, "Method used for token generation"),
  AP_INIT_ITERATE("FileSystems",   dav_dpm_cmd_filesystems, NULL, ACCESS_CONF, "Allowed filesystems to use"),
  {NULL}
};

/**
 * Callback used to register the hooks
 * This is the entry point of the module, where all hooks are registered in
 * mod_dav so the requests can be passed
 * @param p A pool to use for allocations
 */
static void mod_dav_dpm_register_hooks(apr_pool_t *p)
{
  dav_dpm_mime_init(p, "/etc/mime.types");

  dav_hook_find_liveprop(dav_dpm_find_liveprop, NULL, NULL, APR_HOOK_MIDDLE);
  dav_hook_insert_all_liveprops(dav_dpm_insert_all_liveprops, NULL, NULL, APR_HOOK_MIDDLE);
  dav_register_liveprop_group(p, &dav_dpm_liveprop_group);

  dav_register_provider(p, "dpm", &dav_dpm_provider);

  ap_hook_child_init(dav_dpm_child_init, NULL, NULL, APR_HOOK_MIDDLE);

  /* This function has to be called, but it is not thread-safe, so this looks
   * like the best place */
  htext_global_init();
}

/*
 * Module's data structure
 * This is the joint between the module and Apache (callbacks)
 * 'module' is defined in httpd.h
 */
module AP_MODULE_DECLARE_DATA dav_dpm_module =
{
    STANDARD20_MODULE_STUFF,
    dav_dpm_create_dir_config,    /* Per-directory config creator */
    NULL,                         /* Dir config merger            */
    dav_dpm_create_server_config, /* Server config creator        */
    NULL,                         /* Server config merger         */
    dav_dpm_cmds,                 /* Command table                */
    mod_dav_dpm_register_hooks    /* Register hooks               */
};
