/* Bits extracted from ssl_private.h from mod_ssl,
 * which is distributed under the Apache 2 license
 */
#ifndef MOD_SSL_PRIVATE_H
#define	MOD_SSL_PRIVATE_H

#include <http_config.h>
#include <openssl/ssl.h>

/**
 * Provide reasonable defines for some types
 */
#ifndef BOOL
#define BOOL unsigned int
#endif
#ifndef UCHAR
#define UCHAR unsigned char
#endif

/**
 * Define the certificate algorithm types
 */
#define SSL_AIDX_RSA     (0)
#define SSL_AIDX_DSA     (1)
#define SSL_AIDX_MAX     (2)

/**
 * Define the SSL Protocol options
 */
#define SSL_PROTOCOL_NONE  (0)
#define SSL_PROTOCOL_SSLV2 (1<<0)
#define SSL_PROTOCOL_SSLV3 (1<<1)
#define SSL_PROTOCOL_TLSV1 (1<<2)
#define SSL_PROTOCOL_ALL   (SSL_PROTOCOL_SSLV2|SSL_PROTOCOL_SSLV3|SSL_PROTOCOL_TLSV1)
typedef int ssl_proto_t;

/**
 * Define the SSL options
 */
#define SSL_OPT_NONE           (0)
#define SSL_OPT_RELSET         (1<<0)
#define SSL_OPT_STDENVVARS     (1<<1)
#define SSL_OPT_EXPORTCERTDATA (1<<3)
#define SSL_OPT_FAKEBASICAUTH  (1<<4)
#define SSL_OPT_STRICTREQUIRE  (1<<5)
#define SSL_OPT_OPTRENEGOTIATE (1<<6)
#define SSL_OPT_ALL            (SSL_OPT_STDENVVARS|SSL_OPT_EXPORTCERTDATA|SSL_OPT_FAKEBASICAUTH|SSL_OPT_STRICTREQUIRE|SSL_OPT_OPTRENEGOTIATE)
typedef int ssl_opt_t;

/**
 * Define the SSL verify levels
 */
typedef enum {
  SSL_CVERIFY_UNSET           = -1,
  SSL_CVERIFY_NONE            = 0,
  SSL_CVERIFY_OPTIONAL        = 1,
  SSL_CVERIFY_REQUIRE         = 2,
  SSL_CVERIFY_OPTIONAL_NO_CA  = 3
} ssl_verify_t;

/**
 * Define the SSL pass phrase dialog types
 */
typedef enum {
  SSL_PPTYPE_UNSET   = -1,
  SSL_PPTYPE_BUILTIN = 0,
  SSL_PPTYPE_FILTER  = 1,
  SSL_PPTYPE_PIPE    = 2
} ssl_pphrase_t;

/**
 * Define the SSL enabled state
 */
typedef enum {
  SSL_ENABLED_UNSET    = -1,
  SSL_ENABLED_FALSE    = 0,
  SSL_ENABLED_TRUE     = 1,
  SSL_ENABLED_OPTIONAL = 3
} ssl_enabled_t;

/** public cert/private key */
typedef struct {
  /**
   * server only has 1-2 certs/keys
   * 1 RSA and/or 1 DSA
   */
  const char  *cert_files[SSL_AIDX_MAX];
  const char  *key_files[SSL_AIDX_MAX];
  X509        *certs[SSL_AIDX_MAX];
  EVP_PKEY    *keys[SSL_AIDX_MAX];

  /** Certificates which specify the set of CA names which should be
   * sent in the CertificateRequest message: */
  const char  *ca_name_path;
  const char  *ca_name_file;
} modssl_pk_server_t;

/** stuff related to authentication that can also be per-dir */
typedef struct {
  /** known/trusted CAs */
  const char  *ca_cert_path;
  const char  *ca_cert_file;

  const char  *cipher_suite;

  /** for client or downstream server authentication */
  int          verify_depth;
  ssl_verify_t verify_mode;
} modssl_auth_ctx_t;

typedef struct SSLSrvConfigRec SSLSrvConfigRec;

typedef struct {
  SSLSrvConfigRec *sc; /** pointer back to server config */
  SSL_CTX *ssl_ctx;

  /** we are one or the other */
  modssl_pk_server_t *pks;
  void               *pkp;

  ssl_proto_t  protocol;

  /** config for handling encrypted keys */
  ssl_pphrase_t pphrase_dialog_type;
  const char   *pphrase_dialog_path;

  const char  *cert_chain;

  /** certificate revocation list */
  const char  *crl_path;
  const char  *crl_file;
  X509_STORE  *crl;

  modssl_auth_ctx_t auth;
} modssl_ctx_t;

struct SSLSrvConfigRec {
  void            *mc;
  ssl_enabled_t    enabled;
  BOOL             proxy_enabled;
  const char      *vhost_id;
  int              vhost_id_len;
  int              session_cache_timeout;
  BOOL             cipher_server_pref;
  modssl_ctx_t    *server;
  modssl_ctx_t    *proxy;
};

typedef enum {
  SSL_SHUTDOWN_TYPE_UNSET,
  SSL_SHUTDOWN_TYPE_STANDARD,
  SSL_SHUTDOWN_TYPE_UNCLEAN,
  SSL_SHUTDOWN_TYPE_ACCURATE
} ssl_shutdown_type_e;

typedef struct {
  SSL *ssl;
  const char *client_dn;
  X509 *client_cert;
  ssl_shutdown_type_e shutdown_type;
  const char *verify_info;
  const char *verify_error;
  int verify_depth;
  int is_proxy;
  int disabled;
  int non_ssl_request;
} SSLConnRec;

/* Module */
extern module AP_MODULE_DECLARE_DATA ssl_module;

#endif	/* MOD_SSL_PRIVATE_H */

